@@ -10,10 +10,7 @@ use Carp qw(croak);
 use Scalar::Util qw(isweak refaddr reftype);
 
 use KiokuDB::Entry;
-use KiokuDB::Entry::Skip;
 use KiokuDB::Reference;
-use KiokuDB::Collapser::Buffer;
-use KiokuDB::Error::UnknownObjects;
 
 use Data::Visitor 0.18;
 
@@ -23,18 +20,10 @@ use namespace::clean -except => 'meta';
 
 extends qw(Data::Visitor);
 
-with qw(KiokuDB::Role::UUIDs);
-
 has '+tied_as_objects' => ( default => 1 );
 
-has live_objects => (
-    isa => "KiokuDB::LiveObjects",
-    is  => "ro",
-    required => 1,
-);
-
-has backend => (
-    does => "KiokuDB::Backend",
+has resolver => (
+    isa => "KiokuDB::Resolver",
     is  => "ro",
     required => 1,
 );
@@ -42,7 +31,7 @@ has backend => (
 has typemap_resolver => (
     isa => "KiokuDB::TypeMap::Resolver",
     is  => "ro",
-    handles => [qw(collapse_method id_method)],
+    handles => [qw(collapse_method)],
     required => 1,
 );
 
@@ -56,13 +45,76 @@ has '+weaken' => (
     default => 0,
 );
 
-has '_buffer' => (
-    isa => "KiokuDB::Collapser::Buffer",
+has _entries => (
+    isa => 'HashRef',
     is  => "ro",
-    clearer => "_clear_buffer",
-    writer  => "_set_buffer",
+    init_arg => undef,
+    clearer => "_clear_entries",
+    writer  => "_set_entries",
 );
 
+# a list of the IDs of all simple entries
+has _simple_entries => (
+    isa => 'ArrayRef',
+    is  => "ro",
+    init_arg => undef,
+    clearer => "_clear_simple_entries",
+    writer  => "_set_simple_entries",
+);
+
+# keeps track of the simple references which are first class (either weak or
+# shared, and must have an entry)
+has _first_class => (
+    isa => 'HashRef',
+    is  => "ro",
+    init_arg => undef,
+    clearer => "_clear_first_class",
+    writer  => "_set_first_class",
+);
+
+has _options => (
+    isa => 'HashRef',
+    is  => "ro",
+    init_arg => undef,
+    clearer => "_clear_options",
+    writer  => "_set_options",
+);
+
+sub clear_temp_structs {
+    my $self = shift;
+    $self->_clear_entries;
+    $self->_clear_simple_entries;
+    $self->_clear_first_class;
+    $self->_clear_options;
+}
+
+sub collapse_objects {
+    my ( $self, @objects ) = @_;
+
+    my ( $entries, @ids ) = $self->collapse( objects => \@objects );
+
+    # compute the root set
+    my @root_set = delete @{ $entries }{@ids};
+
+    # return the root set and all additional necessary entries
+    return ( @root_set, values %$entries );
+}
+
+sub collapse_known_objects {
+    my ( $self, @objects ) = @_;
+
+    my ( $entries, @ids ) = $self->collapse(
+        objects    => \@objects,
+        only_known => 1,
+    );
+
+    my @root_set = map { $_ and delete $entries->{$_} } @ids;
+
+    # return the root set and all additional necessary entries
+    # may contain undefs
+    return ( @root_set, values %$entries );
+}
+
 sub collapse {
     my ( $self, %args ) = @_;
 
@@ -70,99 +122,127 @@ sub collapse {
 
     my $r;
 
+    if ( $args{only_known} ) {
+        $args{live_objects}  ||= $self->resolver->live_objects;
+        $r = $args{resolver} ||= $args{live_objects};
+    } else {
+        $r = $args{resolver} ||= $self->resolver;
+        $args{live_objects}  ||= $args{resolver}->live_objects;
+    }
+
     if ( $args{shallow} ) {
         $args{only} = set(@$objects);
     }
 
-    my $buf = KiokuDB::Collapser::Buffer->new(
-        live_objects => $self->live_objects,
-        options      => \%args,
-    );
+    my $g = Scope::Guard->new(sub {
+        $self->clear_temp_structs;
+    });
 
-    my $g = Scope::Guard->new(sub { $self->_clear_buffer });
-    $self->_set_buffer($buf);
+    my ( %entries, %fc );
+
+    $self->_set_entries(\%entries);
+    $self->_set_options(\%args);
+    $self->_set_first_class(\%fc);
+    $self->_set_simple_entries([]);
 
     # recurse through the object, accumilating entries
     $self->visit(@$objects);
 
-    my @ids = $buf->merged_objects_to_ids(@$objects);
-
-    $buf->first_class->insert(@ids);
+    my @ids = $args{live_objects}->objects_to_ids(@$objects);
+    @fc{@ids} = ();
 
     # compact UUID space by merging simple non shared structures into a single
     # deep entry
-    $buf->compact_entries if $self->compact;
+    $self->compact_entries() if $self->compact;
 
-    return ( $buf, @ids );
+    return ( \%entries, @ids );
 }
 
-sub may_compact {
-    my ( $self, $ref_or_id ) = @_;
+sub compact_entries {
+    my $self = shift;
 
-    my $id = ref($ref_or_id) ? $ref_or_id->id : $ref_or_id;
+    my ( $entries, $fc, $simple, $options ) = ( $self->_entries, $self->_first_class, $self->_simple_entries, $self->_options );
 
-    not $self->_buffer->first_class->includes($id);
+    # unify non shared simple references
+    if ( my @flatten = grep { not exists $fc->{$_} } @$simple ) {
+        my $flatten = {};
+        @{$flatten}{@flatten} = delete @{$entries}{@flatten};
+
+
+        $options->{resolver}->remove(@flatten);
+
+        $self->compact_entry($_, $flatten) for values %$entries;
+    }
+}
+
+sub compact_entry {
+    my ( $self, $entry, $flatten ) = @_;
+
+    my $data = $entry->data;
+
+    if ( $self->compact_data($data, $flatten) ) {
+        $entry->data($data);
+    }
+}
+
+sub compact_data {
+    my ( $self, $data, $flatten ) = @_;
+
+    if ( ref $data eq 'KiokuDB::Reference' ) {
+        my $id = $data->id;
+
+        if ( my $entry = $flatten->{$id} ) {
+            # replace reference with data from entry, so that the
+            # simple data is inlined, and mark that entry for removal
+            $self->compact_entry($entry, $flatten);
+
+            if ( $entry->tied or $entry->class ) {
+                $entry->clear_id;
+                $_[1] = $entry;
+            } else {
+                $_[1] = $entry->data;
+            }
+            return 1;
+        }
+    } elsif ( ref($data) eq 'ARRAY' ) {
+        ref && $self->compact_data($_, $flatten) for @$data;
+    } elsif ( ref($data) eq 'HASH' ) {
+        ref && $self->compact_data($_, $flatten) for values %$data;
+    } elsif ( ref($data) eq 'KiokuDB::Entry' ) {
+        $self->compact_entry($data, $flatten);
+    } else {
+        # passthrough
+    }
+
+    return;
 }
 
 sub make_entry {
     my ( $self, %args ) = @_;
 
-    my $meta = delete $args{meta} || {};
-
     my $object = $args{object};
 
+    my $live_objects = $self->_options->{live_objects};
+
     if ( my $id = $args{id} ) {
-        my $l = $self->live_objects;
+        my $prev = $live_objects->object_to_entry($object);
 
-        my $prev = $l->object_to_entry($object);
-
-        if ( !$prev and $l->id_in_storage($id) ) {
-            # FIXME Backend->store( insert => [ ... ], update => [ ... ] )
-            # this happens when keep_entries is false
-            $prev = KiokuDB::Entry->new( root => $l->id_in_root_set($id) ); # force the operation to be an update
-        }
-
-        my $entry = KiokuDB::Entry->new(
+        return $self->_entries->{$id} = KiokuDB::Entry->new(
             ( $prev ? ( prev => $prev ) : () ),
             %args,
         );
-
-        $self->_buffer->insert_entry( $id => $entry, $object, %$meta );
-
-        return $entry;
     } else {
         # intrinsic
-        my $entry = KiokuDB::Entry->new(%args);
-
-        $self->_buffer->insert_intrinsic( $object => $entry, %$meta );
-
-        return $entry;
+        return KiokuDB::Entry->new(%args);
     }
 }
 
-sub make_skip_entry {
-    my ( $self, %args ) = @_;
-
-    my $object = $args{object};
-
-    my $prev = $args{prev} || $self->live_objects->object_to_entry($object);
-
-    my $id = $args{id};
-
-    unless ( $id ) {
-        croak "skip entries must have an ID" unless $prev;
-        $id = $prev->id;
-    }
-
-    return undef;
-}
-
 sub make_ref {
     my ( $self, $id, $value ) = @_;
 
     my $weak = isweak($_[2]);
 
-    $self->_buffer->first_class->insert($id) if $weak;
+    $self->_first_class->{$id} = undef if $weak;
 
     return KiokuDB::Reference->new(
         id => $id,
@@ -173,31 +253,35 @@ sub make_ref {
 sub visit_seen {
     my ( $self, $seen, $prev ) = @_;
 
-    my $b = $self->_buffer;
+    my $id = $self->_seen_id($seen);
 
-    if ( my $entry = $b->intrinsic_entry($seen) ) {
-        return $entry->clone;
-    } elsif ( my $id = $self->_buffer->object_to_id($seen) || $self->live_objects->object_to_id($seen) ) {
-        $self->_buffer->first_class->insert($id) unless blessed($seen);
 
-        # return a uuid ref
-        return $self->make_ref( $id => $_[1] );
-    } else {
-        KiokuDB::Error::UnknownObjects->throw( objects => [ $seen ] );
-    }
+    # register ID as first class
+    $self->_first_class->{$id} = undef;
+
+    # return a uuid ref
+    return $self->make_ref( $id => $_[1] );
 }
 
-sub visit_ref_fallback {
+sub _seen_id {
+    my ( $self, $seen ) = @_;
+
+    if ( my $id = $self->_options->{live_objects}->object_to_id($seen) ) {
+        return $id;
+    }
+
+    die { unknown => $seen };
+}
+
+sub visit_ref {
     my ( $self, $ref ) = @_;
 
-    my $o = $self->_buffer->options;
-
-    if ( my $entry = $o->{only_in_storage} && $self->live_objects->object_to_entry($ref) ) {
+    if ( my $entry = $self->_options->{only_new} && $self->_options->{live_objects}->object_to_entry($ref) ) {
         return $self->make_ref( $entry->id => $_[1] );
     }
 
     if ( my $id = $self->_ref_id($ref) ) {
-        if ( !$self->compact and my $only = $o->{only} ) {
+        if ( !$self->compact and my $only = $self->_options->{only} ) {
             unless ( $only->contains($ref) ) {
                 return $self->make_ref( $id => $_[1] );
             }
@@ -208,7 +292,7 @@ sub visit_ref_fallback {
         if ( ref($collapsed) eq 'KiokuDB::Reference' and $collapsed->id eq $id ) {
             return $collapsed; # tied
         } else {
-            push @{ $self->_buffer->simple_entries }, $id;
+            push @{ $self->_simple_entries }, $id;
 
             $self->make_entry(
                 id     => $id,
@@ -223,7 +307,7 @@ sub visit_ref_fallback {
         # compacting it later.
         return $self->SUPER::visit_ref($_[1]);
     } else {
-        KiokuDB::Error::UnknownObjects->throw( objects => [ $ref ] );
+        die { unknown => $ref };
     }
 }
 
@@ -235,27 +319,15 @@ sub visit_ref_data {
 sub _ref_id {
     my ( $self, $ref ) = @_;
 
-    my $l = $self->live_objects;
-
-    if ( my $id = $l->object_to_id($ref) ) {
+    if ( my $id = $self->_options->{resolver}->object_to_id($ref) ) {
         return $id;
-    } else {
-        my $b = $self->_buffer;
-
-        if ( $b->options->{only_known} ) {
-            if ( $self->compact ) {
-                # if we're compacting this is not an error, we just compact in place
-                # and we generate an error if we encounter this data again in visit_seen
-                return;
-            } else {
-                KiokuDB::Error::UnknownObjects->throw( objects => [ $ref ] );
-            }
-        } else {
-            my $id = $self->generate_uuid;
-            $b->insert( $id => $ref );
-            return $id;
-        }
+    } elsif ( $self->compact ) {
+        # if we're compacting this is not an error, we just compact in place
+        # and we generate an error if we encounter this data again in _seen_id
+        return;
     }
+
+    die { unknown => $ref };
 }
 
 # avoid retying, we want to get back Reference or Entry objects
@@ -270,19 +342,19 @@ sub visit_tied {
     my $tie = $self->visit($tied);
 
     if ( my $id = $self->_ref_id($ref) ) {
-        if ( !$self->compact and my $only = $self->_buffer->options->{only} ) {
+        if ( !$self->compact and my $only = $self->_options->{only} ) {
             unless ( $only->contains($ref) ) {
                 return $self->make_ref( $id => $_[1] );
             }
         }
 
-        push @{ $self->_buffer->simple_entries }, $id;
+        push @{ $self->_simple_entries }, $id;
 
         $self->make_entry(
             id     => $id,
             object => $ref,
             data   => $tie,
-            tied   => substr(reftype($ref), 0, 1),
+            tied   => reftype($ref),
         );
 
         return $self->make_ref( $id => $_[2] );
@@ -290,101 +362,85 @@ sub visit_tied {
         return $self->make_entry(
             object => $ref,
             data   => $tie,
-            tied   => substr(reftype($ref), 0, 1),
+            tied   => reftype($ref),
         );
     }
 }
 
-sub visit_object { shift->visit_with_typemap(@_) }
-sub visit_ref { shift->visit_with_typemap(@_) }
+sub visit_object {
+    my ( $self, $object ) = @_;
 
-sub visit_with_typemap {
-    my ( $self, $ref ) = @_;
+    my $collapse = $self->collapse_method(ref $object);
 
-    my $collapse = $self->collapse_method(ref $ref);
-
-    shift->$collapse(@_);
+    $self->$collapse($_[1]);
 }
 
 sub collapse_first_class {
     my ( $self, $collapse, $object, @entry_args ) = @_;
 
+    if ( my $entry = $self->_options->{only_new} && $self->_options->{live_objects}->object_to_entry($object) ) {
+        return $self->make_ref( $entry->id => $_[1] );
+    }
+
     # Data::Visitor stuff for circular refs
     $self->_register_mapping( $object, $object );
 
-    my ( $l, $b ) = ( $self->live_objects, $self->_buffer );
+    my $id = $self->_object_id($object, @entry_args) || return;
 
-    my $id = $l->object_to_id($object);
-    my $in_storage = $l->id_in_storage($id);
-
-    my $o = $b->options;
-
-    if ( $o->{only_in_storage} && $in_storage ) {
-        return $self->make_ref( $id => $_[2] );
-    }
-
-    if ( my $only = $o->{only} ) {
+    if ( my $only = $self->_options->{only} ) {
         unless ( $only->contains($object) ) {
-            if ( $in_storage ) {
-                return $self->make_ref( $id => $_[2] );
-            } else {
-                KiokuDB::Error::UnknownObjects->throw( objects => [ $object ] );
-            }
+            return $self->make_ref( $id => $_[1] );
         }
     }
 
-    unless ( $id ) {
-        if ( $o->{only_known} ) {
-            KiokuDB::Error::UnknownObjects->throw( objects => [ $object ] );
-        } else {
-            my $id_method = $self->id_method(ref $object);
-
-            $id = $self->$id_method($object);
-
-            if ( defined( my $conflict = $l->id_to_object($id) ) ) {
-                return $self->id_conflict( $id, $_[2], $conflict );
-            } else {
-                $b->insert( $id => $object );
-            }
-        }
-    }
+    my $class = ref $object;
 
     my @args = (
         object => $object,
         id     => $id,
-        class  => ref($object),
+        class  => $class,
         @entry_args,
     );
 
-    $self->$collapse(@args);
+    my $data = $self->$collapse(@args);
+
+    $self->make_entry(
+        @args,
+        data => $data,
+    );
 
     # we pass $_[1], an alias, so that isweak works
-    return $self->make_ref( $id => $_[2] );
+    return $self->make_ref( $id => $_[1] );
 }
 
-sub id_conflict {
-    my ( $self, $id, $object, $other ) = @_;
-
-    $self->make_skip_entry( id => $id, object => $object );
-
-    $self->_buffer->insert( $id => $object );
-
-    return $self->make_ref( $id => $_[2] );
-}
-
-
 sub collapse_intrinsic {
     my ( $self, $collapse, $object, @entry_args ) = @_;
 
     my $class = ref $object;
 
+    delete $self->{_seen}{ refaddr($object) }; # FIXME Data::Visitor ->_remove_mapping?
+
     my @args = (
         object => $object,
         class  => $class,
         @entry_args,
     );
 
-    return $self->$collapse(@args);
+    return $self->make_entry(
+        data  => $self->$collapse(@args),
+        class => $class,
+    );
+}
+
+sub _object_id {
+    my ( $self, $object, %args ) = @_;
+
+    if ( defined(my $id = $args{id}) ) {
+        $self->_options->{live_objects}->insert( $id => $object ); # FIXME always?
+        return $id;
+    } else {
+        $self->_options->{resolver}->object_to_id($object) or die { unknown => $object };
+    }
 }
 
 # we don't reblass in collapse_naive
@@ -423,8 +479,7 @@ Object collapsing is detailed in L</"COLLAPSING STRATEGIES">.
 The object's data will be copied into the L<KiokuDB::Entry> with references to
 other data structures translated into L<KiokuDB::Reference> objects.
 
-Reference addresses are mapped to unique identifiers, which are generated as
-necessary.
+Reference addresses are resolved to UIDs by L<KiokuDB::Resolver>.
 
 =head2 Compacting
 
@@ -450,7 +505,7 @@ The resolver consults the typemap (L<KiokuDB::TypeMap>), and caches the results
 as keyed by C<ref $object>.
 
 The typemap contains normal entries (keyed by C<ref $object eq $class>) or isa
-entries (filtered by C<< $object->isa($class) >>). The rationale is that a typemap
+entries (filtered by C<$object->isa($class)>). The rationale is that a typemap
 entry for a superclass might not support all subclasses as well.
 
 Any strategy may be collapsed as a first class object, or intrinsicly, inside
